% Cerca veloce.pl

% Trova delle sequenze di mosse che permettono di spostare solo 3 o 4 spigoli, ricercando in tutto l'albero fino ad una
% profondit prestabilita (con inizia(Profondit)).
% Le mosse possibili riguardano solo la faccia bianca e quella verde.
% Dopo 6 mosse (1 secondo di calcolo): 4 spigoli spostati;
% Dopo 7 (2 s): 4 spigoli;
% Dopo 8 (5 s);
% Dopo 9 (12 s): 4 spigoli;
% Dopo 10 (37 s): 3 spigoli spostati;
% Per ogni mossa ci sono tre possibili scelte, dunque dopo 10 mosse vengono analizzati 120.000 posizioni.
% Trova anche delle sequenze di mosse che permettono di spostare solo 4 vertici. (1.080.000 posizioni).
% Dopo 12 mosse (300 secondi di calcolo): 4 vertici spostati.
% Dopo 16 mosse (6 ore 41 secondi di calcolo): 2 vertici con orientamento spostato. (87.500.000 posizioni).
% Per visualizzare solo le sequenze relative ai vertici  utile modificare la funzione "conta".
% NB: Per esplorare l'albero completo  necessario tutto il tempo, ma le prime soluzioni si hanno molto prima.

% La codifica dei pezzi: [[pezzo,direzione]]:
% b=bianco, v=verde, r=rosso, a=arancio, g=giallo, c=cobalto (blu).
% x direzione bianco-cobalto, y direzione verde-giallo, z direzione rosso-arancio.
% la prima lettera indica il colore rispetto cui  riferita la direzione.
vertici_ok([[bvr,x],[brg,x],[bga,x],[bav,x],[cva,x],[cag,x],[cgr,x],[crv,x]]).
%	        1	       2	     3	   4	  5	6          7         8
spigoli_ok([[bv,x],[br,x],[bg,x],[ba,x],[vr,y],[gr,y],[ga,y],[va,y],[ca,x],[cg,x],[cr,x],[cv,x]]).
%	         1	      2        3        4	        5	    6	7         8	       9       10      11      12

% MOSSA
% mossa(X,Prec,Mossa,Vertici_in,Vertici_out,Spigoli_in,Spigoli_out). 
% X  il nome della faccia che viene ruotata in senso orario, Prec  la faccia mossa precedentemente,
% Mossa indica di quanto ruotare.
mossa(b,Prec,b1,[[V1,D1],[V2,D2],[V3,D3],[V4,D4]|Rest_v],  [[V4,Dd],[V1,Da],[V2,Db],[V3,Dc]|Rest_v],
	[[S1,T1],[S2,T2],[S3,T3],[S4,T4]|Rest_s],  [[S4,Td],[S1,Ta],[S2,Tb],[S3,Tc]|Rest_s]) :-
	Prec\==b,
	cambia_dir(yz,D1,Da),cambia_dir(yz,D2,Db),cambia_dir(yz,D3,Dc),cambia_dir(yz,D4,Dd),
	cambia_dir(yz,T1,Ta),cambia_dir(yz,T2,Tb),cambia_dir(yz,T3,Tc),cambia_dir(yz,T4,Td).
mossa(v,Prec,v1,[[V1,D1],A,B,[V4,D4],[V5,D5],C,D,[V8,D8]],  [[V8,Dd],A,B,[V1,Da],[V4,Db],C,D,[V5,Dc]],
	[[S1,T1],E,F,G,[S5,T5],H,I,[S8,T8],J,K,L,[S12,T12]],  [[S5,Tb],E,F,G,[S12,Td],H,I,[S1,Ta],J,K,L,[S8,Tc]]) :-
	Prec\==v,
	cambia_dir(xz,D1,Da),cambia_dir(xz,D4,Db),cambia_dir(xz,D5,Dc),cambia_dir(xz,D8,Dd),
	cambia_dir(xz,T1,Ta),cambia_dir(xz,T5,Tb),cambia_dir(xz,T8,Tc),cambia_dir(xz,T12,Td).
mossa(b,Prec,b2,[V1,V2,V3,V4|Rest_v],  [V3,V4,V1,V2|Rest_v],
	[S1,S2,S3,S4|Rest_s],  [S3,S4,S1,S2|Rest_s]) :- Prec\==b.
mossa(b,Prec,b3,[[V1,D1],[V2,D2],[V3,D3],[V4,D4]|Rest_v],  [[V2,Db],[V3,Dc],[V4,Dd],[V1,Da]|Rest_v],
	[[S1,T1],[S2,T2],[S3,T3],[S4,T4]|Rest_s],  [[S2,Tb],[S3,Tc],[S4,Td],[S1,Ta]|Rest_s]) :-
	Prec\==b,
	cambia_dir(yz,D1,Da),cambia_dir(yz,D2,Db),cambia_dir(yz,D3,Dc),cambia_dir(yz,D4,Dd),
	cambia_dir(yz,T1,Ta),cambia_dir(yz,T2,Tb),cambia_dir(yz,T3,Tc),cambia_dir(yz,T4,Td).
mossa(v,Prec,v2,[V1,A,B,V4,V5,C,D,V8],  [V5,A,B,V8,V1,C,D,V4],
	[S1,E,F,G,S5,H,I,S8,J,K,L,S12],  [S12,E,F,G,S8,H,I,S5,J,K,L,S1]) :- Prec\==v.
mossa(v,Prec,v3,[[V1,D1],A,B,[V4,D4],[V5,D5],C,D,[V8,D8]],  [[V4,Db],A,B,[V5,Dc],[V8,Dd],C,D,[V1,Da]],
	[[S1,T1],E,F,G,[S5,T5],H,I,[S8,T8],J,K,L,[S12,T12]],  [[S8,Tc],E,F,G,[S1,Ta],H,I,[S12,Td],J,K,L,[S5,Tb]]) :-
	Prec\==v,
	cambia_dir(xz,D1,Da),cambia_dir(xz,D4,Db),cambia_dir(xz,D5,Dc),cambia_dir(xz,D8,Dd),
	cambia_dir(xz,T1,Ta),cambia_dir(xz,T5,Tb),cambia_dir(xz,T8,Tc),cambia_dir(xz,T12,Td).

% CAMBIA_DIR
cambia_dir(xy,z,z):- !. 
cambia_dir(xy,x,y):- !.
cambia_dir(xy,y,x).

cambia_dir(xz,y,y):- !.
cambia_dir(xz,x,z):- !.
cambia_dir(xz,z,x).

cambia_dir(yz,x,x):- !.
cambia_dir(yz,y,z):- !.
cambia_dir(yz,z,y).

% INIZIA:	Da qui si parte
inizia(Numero_mosse):-
assert(sequenza(a,a,a,a,a)), abolish(sequenza,5), assert(sequenza(a,a,a,a,a)), % reinizializza il database
	vertici_ok(V), spigoli_ok(S), trova(z,[],V,S,Numero_mosse,0).

% TROVA:	motore di ricerca
%trova(Precedente,Sequenza,Vertici,Spigoli,Profondita_max,Numero_mosse).
trova(_,_,_,_,N,N):- !, fail. % forzo il backtracking perch voglio trovare tutte le soluzioni
trova(Prec,Sequenza,Vertici,Spigoli,Mosse_max,Numero_mosse):-
	N is Numero_mosse+1,
	mossa(M,Prec,Mossa,Vertici,Vn,Spigoli,Sn),
	conta([Mossa|Sequenza],Vn,Sn,N),
	trova(M,[Mossa|Sequenza],Vn,Sn,Mosse_max,N).
%trova(_,_,_,_,_,_). % Servirebbe per non avere la risposta "no", poi servirebbe un fail...
% NB: per maggiore efficienza  la Sequenza di mosse  dall'ultima alla prima

% CONTA:	controlla se siamo arrivati ad uno stato utile.
conta(Sequenza,Vertici,Spigoli,Numero_mosse):-
	vertici_ok(V_ok), differenze(Vertici,V_ok,0,Nv),
	spigoli_ok(S_ok), differenze(Spigoli,S_ok,0,Ns),
	N is Nv + Ns, N>0, N<5, !,		% numero di pezzi fuori posto
%	Ns=0, Nv<5, N is Nv, !,			% da attivare al posto della linea precedente per  
						% selezionare solo le sequenze per i vertici
	diversi(Vertici,V_ok,L), diversi(Spigoli,S_ok,L1), 	% seleziono solo i pezzi fuori posto
	reverse(Sequenza,Sequenza1), 				% raddrizzo la lista di mosse
	stampa(N,Numero_mosse,Sequenza1,L,L1).
conta(_,_,_,_).

% DIFFERENZE (conta le posizioni non corrette: pi efficiente di diversi)
differenze([],[],N,N).
differenze([H|T],[H1|T1],Acc,N):- H=H1, !, differenze(T,T1,Acc,N).
differenze([_|T],[_|T1],Acc,N):- Acc1 is Acc+1,differenze(T,T1,Acc1,N).

% DIVERSI:	seleziona solo i pezzi fuori posto.
diversi([],[],[]).
diversi([H|T],[H1|T1],L):- H=H1, !, diversi(T,T1,L).
diversi([H|T],[H1|T1],[[H,H1]|L]):- diversi(T,T1,L).

% STAMPA:	stampa le sequenze trovate e le memorizza nel database.
stampa(N,_,_,L,L1):-
	sequenza(N,_,_,L,L1), !.
stampa(N,Numero_mosse,Sequenza,L,L1):-
	assertz(sequenza(N,Numero_mosse,Sequenza,L,L1)), % Serve per memorizzare le sequenze.
	write(N), write(' cambiati dopo '), write(Numero_mosse), write(' mosse:'), nl, write(Sequenza), nl,
	write(L), write(L1), nl, nl.

% VEDI_SEQUENZA: funzione comoda per visualizzare nuovamente le sequenze trovate.
vedi_seq:- vedi_sequenza(0,100).
vedi_seq(End):- End>0, vedi_sequenza(0,End).
vedi_seq(Begin,N):- Begin>0, N>0, vedi_sequenza(Begin,N).

vedi_sequenza(Begin,N):-
	bagof((N1,Numero_mosse,Sequenza,L,L1),sequenza(N1,Numero_mosse,Sequenza,L,L1),S),
	S=[_|T], stampa_sequenza(T,Begin,N).	% ho eliminato la sequenza dummy

% STAMPA_SEQUENZA: utilizzata da Vedi_sequenza.
stampa_sequenza([],_,_):- !.
stampa_sequenza([(N,Numero_mosse,Sequenza,L,L1)|T],0,End):-
	End>0, E is End-1,
	write(N), write(' cambiati dopo '), write(Numero_mosse), write(' mosse:'), nl, write(Sequenza), nl,
	write(L), write(L1), nl, nl,
	stampa_sequenza(T,0,E).
stampa_sequenza([_|T],Begin,End):-
	Begin>0, !, B is Begin-1, stampa_sequenza(T,B,End).
stampa_sequenza(_,_,_).

reverse(L,Lr):- reverse(L,[],Lr).
reverse([],Acc,Acc).
reverse([H|T],Acc,Y):- reverse(T,[H|Acc],Y).

% HELP
user_help:- write('inizia(Numero_mosse): inizia una ricerca'), nl.
help_more:- write('vedi_seq, vedi_seq(Quanti), vedi_seq(Primo,Quanti), help, help_all.'), nl.
help_all:- write('mossa/7, cambia_dir/3, inizia/1, trova/6, conta/4, differenze/4, diversi/3, stampa/5, vedi_sequenza/2,
	vedi_seq/0/1/2, stampa_sequenza/3, reverse/2').


% Ecco alcuni esempi delle sequenze trovate:
%
% 3 cambiati dopo 10 mosse: [v1, b1, v1, b1, v1, b3, v3, b3, v3, b3]	% Tipo questa ne ho trovate 33.
% [][[[br, x], [bv, x]], [[va, z], [br, x]], [[bv, z], [va, y]]]	% Questa la utilizzer nella soluzione del cubo.
%
% 3 cambiati dopo 10 mosse: [b3, v3, b3, v3, b3, v1, b1, v1, b1, v1]	% Come la precedente, ma su spigoli diversi.
% [][[[vr, y], [bv, x]], [[bv, x], [ba, x]], [[ba, z], [vr, y]]]		% Nella Soluzione.
%
% 4 cambiati dopo 6 mosse: [b2, v2, b2, v2, b2, v2]		% Tipo questa ne ho trovata 1; altre con 7, 9 mosse.
% [][[[ba, x], [br, x]], [[br, x], [ba, x]], [[va, y], [vr, y]], [[vr, y], [va, y]]]	% Simmetrica
%
% 4 cambiati dopo 12 mosse: [b1, v3, b3, v1, b1, v3, b3, v1, b1, v3, b3, v1] 	% Tipo questa ne ho trovate 54.
% [[[bav, y], [bvr, x]], [[cva, x], [bga, x]], [[bvr, y], [bav, x]], [[bga, x], [cva, x]]][]	% Nella Soluzione.
% Come la precedente, ma su vertici diversi:
% 4 cambiati dopo 12 mosse: [v1, b3, v3, b1, v1, b3, v3, b1, v1, b3, v3, b1]
% [[[bav, y], [bvr, x]], [[crv, x], [brg, x]], [[bvr, y], [bav, x]], [[brg, x], [crv, x]]][]	% Nella Soluzione.
% Mossa interessante: 4 spigoli sulla stessa faccia.
%4 cambiati dopo 12 mosse: [b2, v1, b1, v3, b1, v3, b3, v2, b3, v2, b2, v1]
%[[[bga, x], [bvr, x]], [[bav, y], [brg, x]], [[bvr, z], [bga, x]], [[brg, y], [bav, x]]][]
%
% 2 cambiati dopo 16 mosse: [b1, v2, b3, v1, b3, v2, b1, v2, b1, v1, b3, v3, b3, v1, b1, v2]	% 30 sequenze simili.
% [[[bvr,y],[bvr,x]],[[bav,y],[bav,x]]][]	% Nella soluzione: utile per orientare i vertici.

% CERCA3.1.pl
% Tentativo di ricerca utilizzando 3 faccie: bianca, verde, rossa: 6 scelte possibili per ogni mossa;
% con profondit 6: 420.000 posizioni (120 s). Eliminando le mosse da 2: 4 scelte per ogni mossa:
% dopo 10: 1.570.000 posizioni (400 s). Ho provato anche con profondit 12: 2 ore circa. Nessuna sequenza trovata.

% CERCA3.2.pl
% Tentativo di ricerca utilizzando 3 faccie: bianca, verde, gialla (sono in linea): Nessun risultato come in "Cerca3.1.pl"
