% CERCA12.pl

% Sfruttando le sequenze di 12 mosse che mi permettono di spostare solo 4 vertici cerco ulteriori sequenze. 
% Nella funzione "stampa" si pu selezionare il numero minimo di mosse nella sequenza:
% Questo serve per scartare tutte le sequenze che portano a posizioni che ho gi trovato.
% Dopo 3 mosse (1 secondi di calcolo): 3 vertici con orientamento cambiato;
% Dopo 4 mosse (2 secondi di calcolo): 3 vertici spostati, sfruttando anche una mossa singola;
% Dopo 5 mosse (3 secondi di calcolo): 3 vertici con orientamento cambiato, sfruttando anche una mossa singola
%										e la sua inversa;
% Dopo 6 mosse (15 secondi di calcolo): 3 vertici spostati, sfruttando anche una mossa singola e la sua inversa;
% Nella funzione "diversi" si pu scegliere se considerare i pezzi solo ruotati.


% La codifica dei pezzi: [[pezzo,direzione]]:
% b=bianco, v=verde, r=rosso, a=arancio, g=giallo, c=cobalto (blu).
% x direzione bianco-cobalto, y direzione verde-giallo, z direzione rosso-arancio.
% la prima lettera indica il colore rispetto cui  riferita la direzione.
vertici_ok([[bvr,x],[brg,x],[bga,x],[bav,x],[cva,x],[cag,x],[cgr,x],[crv,x]]).
%	         1	       2	     3	   4	  5	6          7         8
spigoli_ok([[bv,x],[br,x],[bg,x],[ba,x],[vr,y],[gr,y],[ga,y],[va,y],[ca,x],[cg,x],[cr,x],[cv,x]]).
%	         1	      2        3         4       5	    6	7         8	       9       10      11      12

% MOSSA
% mossa(X,Prec,Mossa,Vertici_in,Vertici_out,Spigoli_in,Spigoli_out). 
% X  il nome della faccia che viene ruotata in senso orario, Prec  la faccia mossa precedentemente,
% Mossa indica di quanto ruotare.
mossa(b,Prec,b1,[[V1,D1],[V2,D2],[V3,D3],[V4,D4]|Rest_v],  [[V4,Dd],[V1,Da],[V2,Db],[V3,Dc]|Rest_v],
	[[S1,T1],[S2,T2],[S3,T3],[S4,T4]|Rest_s],  [[S4,Td],[S1,Ta],[S2,Tb],[S3,Tc]|Rest_s]) :-
	Prec\==b,
	cambia_dir(yz,D1,Da),cambia_dir(yz,D2,Db),cambia_dir(yz,D3,Dc),cambia_dir(yz,D4,Dd),
	cambia_dir(yz,T1,Ta),cambia_dir(yz,T2,Tb),cambia_dir(yz,T3,Tc),cambia_dir(yz,T4,Td).
mossa(v,Prec,v1,[[V1,D1],A,B,[V4,D4],[V5,D5],C,D,[V8,D8]],  [[V8,Dd],A,B,[V1,Da],[V4,Db],C,D,[V5,Dc]],
	[[S1,T1],E,F,G,[S5,T5],H,I,[S8,T8],J,K,L,[S12,T12]],  [[S5,Tb],E,F,G,[S12,Td],H,I,[S1,Ta],J,K,L,[S8,Tc]]) :-
	Prec\==v,
	cambia_dir(xz,D1,Da),cambia_dir(xz,D4,Db),cambia_dir(xz,D5,Dc),cambia_dir(xz,D8,Dd),
	cambia_dir(xz,T1,Ta),cambia_dir(xz,T5,Tb),cambia_dir(xz,T8,Tc),cambia_dir(xz,T12,Td).
mossa(b,Prec,b3,[[V1,D1],[V2,D2],[V3,D3],[V4,D4]|Rest_v],  [[V2,Db],[V3,Dc],[V4,Dd],[V1,Da]|Rest_v],
	[[S1,T1],[S2,T2],[S3,T3],[S4,T4]|Rest_s],  [[S2,Tb],[S3,Tc],[S4,Td],[S1,Ta]|Rest_s]) :-
	Prec\==b,
	cambia_dir(yz,D1,Da),cambia_dir(yz,D2,Db),cambia_dir(yz,D3,Dc),cambia_dir(yz,D4,Dd),
	cambia_dir(yz,T1,Ta),cambia_dir(yz,T2,Tb),cambia_dir(yz,T3,Tc),cambia_dir(yz,T4,Td).
mossa(v,Prec,v3,[[V1,D1],A,B,[V4,D4],[V5,D5],C,D,[V8,D8]],  [[V4,Db],A,B,[V5,Dc],[V8,Dd],C,D,[V1,Da]],
	[[S1,T1],E,F,G,[S5,T5],H,I,[S8,T8],J,K,L,[S12,T12]],  [[S8,Tc],E,F,G,[S1,Ta],H,I,[S12,Td],J,K,L,[S5,Tb]]) :-
	Prec\==v,
	cambia_dir(xz,D1,Da),cambia_dir(xz,D4,Db),cambia_dir(xz,D5,Dc),cambia_dir(xz,D8,Dd),
	cambia_dir(xz,T1,Ta),cambia_dir(xz,T5,Tb),cambia_dir(xz,T8,Tc),cambia_dir(xz,T12,Td).

% Sequenze di 12 mosse tramite cui si modificano solo 4 vertici. 
mossa(r,Prec,rv12,[[V1,D1],V2,A,V4,B,C,D,[V8,D8]],  [[V8,Dd],V4,A,V2,B,C,D,[V1,Da]] ,S,S):-
	Prec\==r, cambia_dir(yz,D1,Da), cambia_dir(yz,D8,Dd).
mossa(v,Prec,vb12,[[V1,D1],V2,A,[V4,D4],B,C,D,V8],  [[V4,Dc],V8,A,[V1,Da],B,C,D,V2] ,S,S):-
	Prec\==v, cambia_dir(xy,D1,Da), cambia_dir(xy,D4,Dc).
mossa(b,Prec,br12,[[V1,D1],[V2,D2],A,V4,B,C,D,V8],  [[V2,Db],[V1,Da],A,V8,B,C,D,V4] ,S,S):-
	Prec\==b, cambia_dir(xz,D1,Da), cambia_dir(xz,D2,Db).

% CAMBIA_DIR
cambia_dir(xy,z,z):- !. 
cambia_dir(xy,x,y):- !.
cambia_dir(xy,y,x).

cambia_dir(xz,y,y):- !.
cambia_dir(xz,x,z):- !.
cambia_dir(xz,z,x).

cambia_dir(yz,x,x):- !.
cambia_dir(yz,y,z):- !.
cambia_dir(yz,z,y).

% INIZIA:	Da qui si parte.
inizia(Numero_mosse):- assert(sequenza(a,a,a,a,a)), abolish(sequenza,5), assert(sequenza(a,a,a,a,a)),
 vertici_ok(V), spigoli_ok(S), trova(z,[],V,S,Numero_mosse,0).

% TROVA:	motore di ricerca
% trova(Precedente,Sequenza,Vertici,Spigoli,Profondita_max,Numero_mosse).
trova(_,_,_,_,N,N):- !, fail. % forzo il backtracking perch voglio trovare tutte le soluzioni
trova(Prec,Sequenza,Vertici, Spigoli,Mosse_max,Numero_mosse):-
	N is Numero_mosse+1,
	mossa(M,Prec,Mossa,Vertici,Vn, Spigoli,Sn),
	conta([Mossa|Sequenza],Vn, Sn,N),
	trova(M,[Mossa|Sequenza],Vn, Sn,Mosse_max,N).
%trova(_,_,_,_,_,_). % Servirebbe per non avere la risposta "no", poi servirebbe un fail...
% NB: per maggiore efficienza la Sequenza di mosse  dall'ultima alla prima

% CONTA:	controlla se siamo arrivati ad uno stato utile.
conta(Sequenza,Vertici, Spigoli,Numero_mosse):-
	vertici_ok(V_ok), differenze(Vertici,V_ok,0,Nv),
	spigoli_ok(S_ok), differenze(Spigoli,S_ok,0,Ns),
	N is Nv + Ns, N>0, N<4, !,				% numero di pezzi fuori posto
	diversi(Vertici,V_ok,L),			 		% seleziono solo i pezzi fuori posto
	reverse(Sequenza,Sequenza1), diversi(Spigoli,S_ok,L1),	% raddrizzo la lista di mosse
	stampa(N,Numero_mosse,Sequenza1,L,L1).
conta(_,_,_,_).

% DIFFERENZE:	conta le posizioni non corrette: pi efficiente di diversi.
differenze([],[],N,N).
differenze([H|T],[H1|T1],Acc,N):- H=H1, !, differenze(T,T1,Acc,N).
differenze([_|T],[_|T1],Acc,N):- Acc1 is Acc+1,differenze(T,T1,Acc1,N).

% DIVERSI:	seleziona solo i pezzi fuori posto.
diversi([],[],[]).
diversi([H|T],[H1|T1],L):- H=H1, !, diversi(T,T1,L).
%diversi([[A|_]|_],[[A|_]|_],_):- !, fail.  % Per non considerare i pezzi solo ruotati 
diversi([H|T],[H1|T1],[[H,H1]|L]):- diversi(T,T1,L).

% STAMPA:	stampa le sequenze trovate e le memorizza nel database.
stampa(N,_,_,L,L1):-
	sequenza(N,_,_,L,L1), !.
stampa(N,Numero_mosse,Sequenza,L,L1):-
	assertz(sequenza(N,Numero_mosse,Sequenza,L,L1)),
	Numero_mosse>2,		% per selezionare le sequenze che mi interessano.
	write(N), write(' cambiati dopo '), write(Numero_mosse), write(' mosse:'), nl, write(Sequenza), nl,
	write(L), write(L1), nl, nl.

% VEDI_SEQUENZA:	utilizzata da Vedi_sequenza.
vedi_seq:- vedi_sequenza(0,100).
vedi_seq(End):- End>0, vedi_sequenza(0,End).
vedi_seq(Begin,N):- Begin>0, N>0, vedi_sequenza(Begin,N).

vedi_sequenza(Begin,N):-
	bagof((N1,Numero_mosse,Sequenza,L,L1),sequenza(N1,Numero_mosse,Sequenza,L,L1),S),
	S=[_|T], stampa_sequenza(T,Begin,N).	% ho eliminato la sequenza dummy

% STAMPA_SEQUENZA:	funzione comoda per visualizzare nuovamente le sequenze trovate.
stampa_sequenza([],_,_):- !.
stampa_sequenza([(N,Numero_mosse,Sequenza,L,L1)|T],0,End):-
	End>0, E is End-1,
	write(N), write(' cambiati dopo '), write(Numero_mosse), write(' mosse:'), nl, write(Sequenza), nl,
	write(L), write(L1), nl, nl,
	stampa_sequenza(T,0,E).
stampa_sequenza([_|T],Begin,End):-
	Begin>0, !, B is Begin-1, stampa_sequenza(T,B,End).
stampa_sequenza(_,_,_).

reverse(L,Lr):- reverse(L,[],Lr).
reverse([],Acc,Acc).
reverse([H|T],Acc,Y):- reverse(T,[H|Acc],Y).

% HELP
user_help:- write('inizia(Numero_mosse): inizia una ricerca'), nl.
help_more:- write('vedi_seq, vedi_seq(Quanti), vedi_seq(Primo,Quanti), help, help_all.'), nl.
help_all:- write('mossa/5, cambia_dir/3, inizia/1, trova/5, conta/3, differenze/4, diversi/3, stampa/5, vedi_seq/0/1/2,
	vedi_sequenza/2, stampa_sequenza/3, reverse/2').

% Ecco alcuni esempi delle sequenze trovate:
%
% 3 cambiati dopo 3 mosse: [rv12, vb12, br12]			% 6 sequenze  sequenze simili.
% [[[bvr, z], [bvr, x]], [[brg, y], [brg, x]], [[crv, y], [crv, x]]][]	% Solo orientamenti.
%
% 3 cambiati dopo 4 mosse: [b1, vb12, b3, rv12]	% Questa implicitamente la utilizzo nella soluzione del cubo.
%[[[bga, z], [brg, x]], [[bav, z], [bga, x]], [[brg, x], [bav, x]]][] 		%  8 sequenze simili.
%  importante notare che variano solo i vertici, pur utilizzando 2 mosse singole (si annullano tra loro).
%
% 3 cambiati dopo 5 mosse: [b1, rv12, br12, vb12, b3]	% 14 simili.
% [[[bga, y], [bga, x]], [[bav, z], [bav, x]], [[crv, z], [crv, x]]][]
%
% 3 cambiati dopo 6 mosse: [v1, b1, vb12, b3, rv12, v3]	% Utilizzata implicitamente nella risoluzione.
% [[[brg, z], [bvr, x]], [[bga, z], [brg, x]], [[bvr, x], [bga, x]]][]	% 32 simili.
% Mi  stato di notevole aiuto vedere come con l'aiuto delle mosse singole eseguite in coppia (prima una e dopo
% qualche  sequenza da 12 la sua opposta) sia possibile trovare tante posizioni differenti per i vertici, senza spostare
% nessuno spigolo!
%
% 2 cambiati dopo 6 mosse: [rv12, vb12, rv12, br12, vb12, br12]	% 12 simili.
% [[[bvr, z], [bvr, x]], [[bav, z], [bav, x]]][]
%

