% Risolvi.0.pl (e Risolvi1.pl)

% A partire da una configurazione in cui tutti i pezzi sono fuori posto, cerco di arrivare ad avere tutti i vertici a posto.
% Posso utilizzare 9 sequenze da 12.
% Si sceglie la profondit con "inizia(N)"; con N=5, in 8 secondi arrivo a 2 configurazioni con solo 4 pezzi spostati.
% Con N=7, dopo 6 minuti anche configurazioni con 3 pezzi fuori posto. 
% ( Risolvi1.pl  simile, ma con pi mosse e qualche leggera modifica; arriva alla soluzione in 50 minuti).

% [[pezzo,direzione]]:
% b=bianco, v=verde, r=rosso, a=arancio, g=giallo, c=cobalto (blu).
% x direzione bianco-cobalto, y direzione verde-giallo, z direzione rosso-arancio.
% la prima lettera indica il colore rispetto cui  riferita la direzione.
vertici_ok([[bvr,x],[brg,x],[bga,x],[bav,x],[cva,x],[cag,x],[cgr,x],[crv,x]]).
%	         1	       2	     3	   4	  5	6          7         8
spigoli_ok([[bv,x],[br,x],[bg,x],[ba,x],[vr,y],[gr,y],[ga,y],[va,y],[ca,x],[cg,x],[cr,x],[cv,x]]).
%	         1	     2        3         4	        5	    6	7        8	       9       10      11      12

% MOSSA
% Sequenze di mosse da 12: solo i vertici vengono spostati.
% Le prime tre agiscono tutte sugli stessi 4 verttici.
mossa(r,Prec,rv12,[[V1,D1],V2,A,V4,B,C,D,[V8,D8]],  [[V8,Dd],V4,A,V2,B,C,D,[V1,Da]]):-
	Prec\==rv12, cambia_dir(yz,D1,Da), cambia_dir(yz,D8,Dd).
mossa(v,Prec,vb12,[[V1,D1],V2,A,[V4,D4],B,C,D,V8],  [[V4,Dc],V8,A,[V1,Da],B,C,D,V2]):-
	Prec\==vb12, cambia_dir(xy,D1,Da), cambia_dir(xy,D4,Dc).
mossa(b,Prec,br12,[[V1,D1],[V2,D2],A,V4,B,C,D,V8],  [[V2,Db],[V1,Da],A,V8,B,C,D,V4]):-
	Prec\==br12, cambia_dir(xz,D1,Da), cambia_dir(xz,D2,Db).
% Le successive tre agiscono sugli altri 4 vertici,
mossa(g,Prec,ga12,[A,B,[V3,D3],C,V5,[V6,D6],V7,D],  [A,B,[V6,Dd],C,V7,[V3,Da],V5,D]):-
	Prec\==ga12, cambia_dir(yz,D3,Da), cambia_dir(yz,D6,Dd).
mossa(a,Prec,ac12,[A,B,V3,C,[V5,D5],[V6,D6],V7,D],  [A,B,V7,C,[V6,Dc],[V5,Da],V3,D]):-
	Prec\==ac12, cambia_dir(xz,D5,Da), cambia_dir(xz,D6,Dc).
mossa(c,Prec,cg12,[A,B,V3,C,V5,[V6,D6],[V7,D7],D],  [A,B,V5,C,V3,[V7,Db],[V6,Da],D]):-
	Prec\==cg12, cambia_dir(xy,D6,Da), cambia_dir(xy,D7,Db).
% Tre mosse di collegamento tra i due blocchi
mossa(b,Prec,bg12,[V1,[V2,D2],[V3,D3],A,B,C,V7,D],  [V7,[V3,Db],[V2,Da],A,B,C,V1,D]):-
	Prec\==bg12, cambia_dir(xy,D2,Da), cambia_dir(xy,D3,Db).
mossa(c,Prec,ca12,[A,B,C,V4,[V5,D5],[V6,D6],D,V8],  [A,B,C,V8,[V6,Db],[V5,Da],D,V4]):-
	Prec\==ca12, cambia_dir(xy,D5,Da), cambia_dir(xy,D6,Db).
mossa(v,Prec,vr12,[[V1,D1],A,B,V5,C,D,V7,[V8,D8]],  [[V8,Dd],A,B,V7,C,D,V5,[V1,Da]]):-
	Prec\==rv12, cambia_dir(yz,D1,Da), cambia_dir(yz,D8,Dd).

% CAMBIA_DIR
cambia_dir(xy,z,z):- !. 
cambia_dir(xy,x,y):- !.
cambia_dir(xy,y,x).
cambia_dir(xz,y,y):- !.
cambia_dir(xz,x,z):- !.
cambia_dir(xz,z,x).
cambia_dir(yz,x,x):- !.
cambia_dir(yz,y,z):- !.
cambia_dir(yz,z,y).

% INIZIA:	parto da una configurazione in cui tutti i vertici sono fuori posto.
inizia(Numero_mosse):- assert(sequenza(a,a,a,a)), abolish(sequenza,4), assert(sequenza(a,a,a,a)),
  trova(z, [], [[cgr,z],[bav,x],[cva,y],[cag,y],[brg,y],[crv,y],[bvr,z],[bga,y]], Numero_mosse, 0).

% TROVA
% trova(Precedente,Sequenza,Sequenza_new,Vertici,Profondita_max,Numero_mosse).
trova(_,_,_,N,N):- !, fail. % forzo il backtracking perch voglio trovare tutte le soluzioni
trova(Prec,Sequenza,Vertici,Mosse_max,Numero_mosse):-
	N is Numero_mosse+1,
	mossa(_,Prec,Mossa,Vertici,Vn),
	conta([Mossa|Sequenza],Vn,N),
	trova(Mossa,[Mossa|Sequenza],Vn,Mosse_max,N).
%trova(_,_,_,_,_). % Servirebbe per non avere la risposta "no", poi servirebbe un fail...
% NB: per maggiore efficienza la Sequenza di mosse  dall'ultima alla prima

% CONTA
conta(Sequenza,Vertici,Numero_mosse):- vertici_ok(V_ok), differenze(Vertici,V_ok,0,Nv),
	N is Nv,  N<5, !,						% numero di pezzi fuori posto
	diversi(Vertici,V_ok,L),		 			% seleziono solo i pezzi fuori posto
	reverse(Sequenza,Sequenza1), 				% raddrizzo le lista di mosse
	stampa(N,Numero_mosse,Sequenza1,L).
conta(_,_,_).

% DIVERSI
diversi([],[],[]).
diversi([H|T],[H1|T1],L):- H=H1, !, diversi(T,T1,L).
diversi([H|T],[H1|T1],[[H,H1]|L]):- diversi(T,T1,L).

% DIFFERENZE (pi efficiente di diversi)
differenze([],[],N,N).
differenze([H|T],[H1|T1],Acc,N):- H=H1, !, differenze(T,T1,Acc,N).
differenze([_|T],[_|T1],Acc,N):- Acc1 is Acc+1,differenze(T,T1,Acc1,N).

% STAMPA
stampa(N,_,_,L):- sequenza(N,_,_,L), !.
stampa(N,Numero_mosse,Sequenza,L):- assertz(sequenza(N,Numero_mosse,Sequenza,L)),
	write(N), write(' cambiati dopo '), write(Numero_mosse), write(' mosse:'), nl, write(Sequenza), nl,
	write(L), nl, nl.

% VEDI_SEQUENZA
vedi_seq:- vedi_sequenza(0,100).
vedi_seq(End):- End>0, vedi_sequenza(0,End).
vedi_seq(Begin,N):- Begin>0, N>0, vedi_sequenza(Begin,N).

vedi_sequenza(Begin,N):-
	bagof((N1,Numero_mosse,Sequenza,L),sequenza(N1,Numero_mosse,Sequenza,L),S),
	S=[_|T], stampa_sequenza(T,Begin,N).	% ho eliminato la sequenza dummy

% STAMPA_SEQUENZA
stampa_sequenza([],_,_):- !.
stampa_sequenza([(N,Numero_mosse,Sequenza,L)|T],0,End):- End>0, E is End-1,
	write(N), write(' cambiati dopo '), write(Numero_mosse), write(' mosse:'), nl, write(Sequenza), nl,
	write(L), nl, nl, stampa_sequenza(T,0,E).
stampa_sequenza([_|T],Begin,End):- Begin>0, !, B is Begin-1, stampa_sequenza(T,B,End).
stampa_sequenza(_,_,_).

reverse(L,Lr):- reverse(L,[],Lr).
reverse([],Acc,Acc).
reverse([H|T],Acc,Y):- reverse(T,[H|Acc],Y).

% HELP
user_help:- write('inizia(Numero_mosse): inizia una ricerca'), nl, write('help_more, help_all. '), nl.
help_more:- write('vedi_seq, vedi_seq(Quanti), vedi_seq(Primo,Quanti), help, help_all.'), nl.
help_all:- write('mossa/5, cambia_dir/3, trova/6, conta/4, stampa/5, vedi_seq/0/1/2, vedi_sequenza/2,
stampa_sequenza/3, inizia/1, diversi/3, differenze/4, reverse/2').
