% Vertici.pl

% Programma che partendo da una configurazione in cui tutti i pezzi sono fuori posto riesce a sistemare i vertici 
% lasciando gli spigoli nella posizione originale. L'algoritmo  diviso in due parti:
% 1) posizionare i vertici. 2) sistemare gli orientamenti.
% Per posizionare i vertici utilizzo 6 sequenze di 12 mosse che invertono 4 vertici a 2 a 2, pi 6 mosse singole:
% 3 relative alla faccia bianca, 3 a quella cobalto. Esistono 24 diverse sequenze da 12 e 18 mosse singole, ma quelle che
% ho scelto riescono a sostituire quelle che ho scartato, inoltre con la minima ridondanza. Tempo di calcolo: circa 2 sec.
%  importante notare che le mosse singole vengono sempre effettuate in coppie, una linversa dellaltra e tra le due una 
% o pi sequenze da 12 mosse; tutto ci lascia inalterate le posizioni degli spigoli, ma permette di utilizzare una stessa
% sequenza da 12 su diverse quadruple di vertici.  in questo modo che ho potuto evitare di usare tutte le possibili 
% sequenze da 12.
% In precedenza avevo cercato di posizionare i vertici solo con le sequenze da 12; arrivavo al goal in poco meno di
% un'ora. Poi ho fatto diverse prove (cambiando: il set di mosse che utilizzava l'algoritmo, il numero di vertici che
% dovevano essere posizionati per ricominciare la ricerca da tale punto) ottenendo tempi di soluzione diversi.

% Per orientare i vertici ho provato due algoritmi diversi:
% 1) I vertici vengono orientati usando sequenze di 16 mosse che cambiano orientamento a 2 vertici.
% 2) Come sopra, ma i vertici vengono orientati in un ordine prestabilito e non si fa una ricerca.
% Il  2 metodo  pi veloce, ma alla fine possono essere necessarie pi mosse; per ottenere meno mosse dal primo
% sarebbero necessarie tante sequenze (2*8*7/2=56 in totale).  comunque ininfluente il tempo per orientare i vertici.


% [[pezzo,direzione]]:
% b=bianco, v=verde, r=rosso, a=arancio, g=giallo, c=cobalto (blu).
% x direzione bianco-cobalto, y direzione verde-giallo, z direzione rosso-arancio.
% la prima lettera indica il colore rispetto cui  riferita la direzione.
vertici_ok([[bvr,x],[brg,x],[bga,x],[bav,x],[cva,x],[cag,x],[cgr,x],[crv,x]]).
%	         1	       2	     3	   4	 5	6          7          8
spigoli_ok([[bv,x],[br,x],[bg,x],[ba,x],[vr,y],[gr,y],[ga,y],[va,y],[ca,x],[cg,x],[cr,x],[cv,x]]).
%	         1	      2        3        4	        5	    6	7        8	       9       10      11      12

% Mosse singole per spostare i vertici (non vengono considerati gli spigoli, visto che alla fine tornano a posto) 
mossa_vs(Prec,b1,[[V1,D1],[V2,D2],[V3,D3],[V4,D4]|Rest_v],  [[V4,Dd],[V1,Da],[V2,Db],[V3,Dc]|Rest_v]) :-
	(Prec=b3), cambia_dir(yz,D1,Da),cambia_dir(yz,D2,Db),cambia_dir(yz,D3,Dc),cambia_dir(yz,D4,Dd).
%mossa_vs(Prec,v1,[[V1,D1],A,B,[V4,D4],[V5,D5],C,D,[V8,D8]],  [[V8,Dd],A,B,[V1,Da],[V4,Db],C,D,[V5,Dc]]) :-
%	(Prec=v3), cambia_dir(xz,D1,Da),cambia_dir(xz,D4,Db),cambia_dir(xz,D5,Dc),cambia_dir(xz,D8,Dd).
%mossa_vs(Prec,r1,[[V1,D1],[V2,D2],A,B,C,D,[V7,D7],[V8,D8]],  [[V2,Db],[V7,Dc],A,B,C,D,[V8,Dd],[V1,Da]]) :-
%	(Prec=r3), cambia_dir(xy,D1,Da),cambia_dir(xy,D2,Db),cambia_dir(xy,D7,Dc),cambia_dir(xy,D8,Dd).
%mossa_vs(Prec,g1,[A,[V2,D2],[V3,D3],B,C,[V6,D6],[V7,D7],D],  [A,[V3,Db],[V6,Dc],B,C,[V7,Dd],[V2,Da],D]) :-
%	(Prec=g3), cambia_dir(xz,D2,Da),cambia_dir(xz,D3,Db),cambia_dir(xz,D6,Dc),cambia_dir(xz,D7,Dd).
%mossa_vs(Prec,a1,[A,B,[V3,D3],[V4,D4],[V5,D5],[V6,D6]|Rest_v],  
%	[A,B,[V4,Db],[V5,Dc],[V6,Dd],[V3,Da]|Rest_v]) :-
%	(Prec=a3), cambia_dir(xy,D3,Da),cambia_dir(xy,D4,Db),cambia_dir(xy,D5,Dc),cambia_dir(xy,D6,Dd).
mossa_vs(Prec,c1,[A,B,C,D,[V5,D5],[V6,D6],[V7,D7],[V8,D8]],  [A,B,C,D,[V8,Dd],[V5,Da],[V6,Db],[V7,Dc]]) :-
	(Prec=c3), cambia_dir(yz,D5,Da),cambia_dir(yz,D6,Db),cambia_dir(yz,D7,Dc),cambia_dir(yz,D8,Dd).
mossa_vs(Prec,b3,[[V1,D1],[V2,D2],[V3,D3],[V4,D4]|Rest_v],  [[V2,Db],[V3,Dc],[V4,Dd],[V1,Da]|Rest_v]) :-
	(Prec=b1), cambia_dir(yz,D1,Da),cambia_dir(yz,D2,Db),cambia_dir(yz,D3,Dc),cambia_dir(yz,D4,Dd).
%mossa_vs(Prec,v3,[[V1,D1],A,B,[V4,D4],[V5,D5],C,D,[V8,D8]],  [[V4,Db],A,B,[V5,Dc],[V8,Dd],C,D,[V1,Da]]) :-
%	(Prec=v1), cambia_dir(xz,D1,Da),cambia_dir(xz,D4,Db),cambia_dir(xz,D5,Dc),cambia_dir(xz,D8,Dd).
%mossa_vs(Prec,r3,[[V1,D1],[V2,D2],A,B,C,D,[V7,D7],[V8,D8]],  [[V8,Dd],[V1,Da],A,B,C,D,[V2,Db],[V7,Dc]]) :-
%	(Prec=r1),	cambia_dir(xy,D1,Da),cambia_dir(xy,D2,Db),cambia_dir(xy,D7,Dc),cambia_dir(xy,D8,Dd).
%mossa_vs(Prec,g3,[A,[V2,D2],[V3,D3],B,C,[V6,D6],[V7,D7],D],  [A,[V7,Dd],[V2,Da],B,C,[V3,Db],[V6,Dc],D]) :-
%	(Prec=g1),	cambia_dir(xz,D2,Da),cambia_dir(xz,D3,Db),cambia_dir(xz,D6,Dc),cambia_dir(xz,D7,Dd).
%mossa_vs(Prec,a3,[A,B,[V3,D3],[V4,D4],[V5,D5],[V6,D6]|Rest_v],  
%	[A,B,[V6,Dd],[V3,Da],[V4,Db],[V5,Dc]|Rest_v]) :-
%	(Prec=a1), cambia_dir(xy,D3,Da),cambia_dir(xy,D4,Db),cambia_dir(xy,D5,Dc),cambia_dir(xy,D6,Dd).
mossa_vs(Prec,c3,[A,B,C,D,[V5,D5],[V6,D6],[V7,D7],[V8,D8]],  [A,B,C,D,[V6,Db],[V7,Dc],[V8,Dd],[V5,Da]]) :-
	(Prec=c1), cambia_dir(yz,D5,Da),cambia_dir(yz,D6,Db),cambia_dir(yz,D7,Dc),cambia_dir(yz,D8,Dd).
mossa_vs(Prec,b2,[V1,V2,V3,V4|Rest_v],  [V3,V4,V1,V2|Rest_v]) :- (Prec=b2).
%mossa_vs(Prec,v2,[V1,A,B,V4,V5,C,D,V8],  [V5,A,B,V8,V1,C,D,V4]) :- (Prec=v2).
%mossa_vs(Prec,r2,[V1,V2,A,B,C,D,V7,V8],  [V7,V8,A,B,C,D,V1,V2]) :- (Prec=r2).
%mossa_vs(Prec,g2,[A,V2,V3,B,C,V6,V7,D],  [A,V6,V7,B,C,V2,V3,D]) :- (Prec=r2).
%mossa_vs(Prec,a2,[A,B,V3,V4,V5,V6|Rest_v],  [A,B,V5,V6,V3,V4|Rest_v]) :- (Prec=a2).
mossa_vs(Prec,c2,[A,B,C,D,V5,V6,V7,V8],  [A,B,C,D,V7,V8,V5,V6]) :- (Prec=c2).

% Sequenze di dodici mosse che spostano 4 vertici. Ogni gruppo di tre sequenze lavora sugli stessi 4 vertici.
mossa_v(Prec,rv12,[[V1,D1],V2,A,V4,B,C,D,[V8,D8]],  [[V8,Dd],V4,A,V2,B,C,D,[V1,Da]]):-
	Prec\==rv12, cambia_dir(yz,D1,Da), cambia_dir(yz,D8,Dd). % Sequenza: [r,v3,r3,v,r,v3,r3,v,r,v3,r3,v]
mossa_v(Prec,vb12,[[V1,D1],V2,A,[V4,D4],B,C,D,V8],  [[V4,Dc],V8,A,[V1,Da],B,C,D,V2]):-
	Prec\==vb12, cambia_dir(xy,D1,Da), cambia_dir(xy,D4,Dc).
mossa_v(Prec,br12,[[V1,D1],[V2,D2],A,V4,B,C,D,V8],  [[V2,Db],[V1,Da],A,V8,B,C,D,V4]):-
	Prec\==br12, cambia_dir(xz,D1,Da), cambia_dir(xz,D2,Db).

%mossa_v(Prec,bg12,[V1,[V2,D2],[V3,D3],A,B,C,V7,D],  [V7,[V3,Db],[V2,Da],A,B,C,V1,D]):-
%	Prec\==bg12, cambia_dir(xy,D2,Da), cambia_dir(xy,D3,Db).
%mossa_v(Prec,rb12,[[V1,D1],[V2,D2],V3,A,B,C,V7,D],  [[V2,Db],[V1,Da],V7,A,B,C,V3,D]):-
%	Prec\==rb12, cambia_dir(xz,D1,Da), cambia_dir(xz,D2,Db).
%mossa_v(Prec,gr12,[V1,[V2,D2],V3,A,B,C,[V7,D7],D],  [V3,[V7,Db],V1,A,B,C,[V2,Da],D]):-
%	Prec\==gr12, cambia_dir(yz,D2,Da), cambia_dir(yz,D7,Db).

%mossa_v(Prec,ba12,[A,V2,[V3,D3],[V4,D4],B,V6,C,D],  [A,V6,[V4,Db],[V3,Da],B,V2,C,D]):-
%	Prec\==ba12, cambia_dir(xz,D3,Da), cambia_dir(xz,D4,Db).
%mossa_v(Prec,ag12,[A,V2,[V3,D3],V4,B,[V6,D6],C,D],  [A,V4,[V6,Db],V2,B,[V3,Da],C,D]):-
%	Prec\==ag12, cambia_dir(yz,D3,Da), cambia_dir(yz,D6,Db).
%mossa_v(Prec,gb12,[A,V2,[V3,D3],[V4,D4],B,V6,C,D],  [A,V6,[V4,Db],[V3,Da],B,V2,C,D]):-
%	Prec\==gb12, cambia_dir(xy,D3,Da), cambia_dir(xy,D4,Db).

%mossa_v(Prec,bv12,[[V1,D1],A,V3,[V4,D4],V5,B,C,D],  [[V4,Db],A,V5,[V1,Da],V3,B,C,D]):-
%	Prec\==bv12, cambia_dir(xy,D1,Da), cambia_dir(xy,D4,Db).
%mossa_v(Prec,va12,[V1,A,V3,[V4,D4],[V5,D5],B,C,D],  [V3,A,V1,[V5,Db],[V4,Da],B,C,D]):-
%	Prec\==va12, cambia_dir(yz,D4,Da), cambia_dir(yz,D5,Db).
%mossa_v(Prec,ab12,[V1,A,[V3,D3],[V4,D4],V5,B,C,D],  [V5,A,[V4,Db],[V3,Da],V1,B,C,D]):-
%	Prec\==ab12, cambia_dir(xz,D3,Da), cambia_dir(xz,D4,Db).

%mossa_v(Prec,ca12,[A,B,C,V4,[V5,D5],[V6,D6],D,V8],  [A,B,C,V8,[V6,Db],[V5,Da],D,V4]):-
%	Prec\==ca12, cambia_dir(xz,D5,Da), cambia_dir(xz,D6,Db).
%mossa_v(Prec,av12,[A,B,C,[V4,D4],[V5,D5],V6,D,V8],  [A,B,C,[V5,Db],[V4,Da],V8,D,V6]):-
%	Prec\==av12, cambia_dir(yz,D4,Da), cambia_dir(yz,D5,Db).
%mossa_v(Prec,vc12,[A,B,C,V4,[V5,D5],V6,D,[V8,D8]],  [A,B,C,V6,[V8,Db],V4,D,[V5,Da]]):-
%	Prec\==vc12, cambia_dir(xy,D5,Da), cambia_dir(xy,D8,Db).

mossa_v(Prec,ga12,[A,B,[V3,D3],C,V5,[V6,D6],V7,D],  [A,B,[V6,Dd],C,V7,[V3,Da],V5,D]):-
	Prec\==ga12, cambia_dir(yz,D3,Da), cambia_dir(yz,D6,Dd).
mossa_v(Prec,ac12,[A,B,V3,C,[V5,D5],[V6,D6],V7,D],  [A,B,V7,C,[V6,Dc],[V5,Da],V3,D]):-
	Prec\==ac12, cambia_dir(xz,D5,Da), cambia_dir(xz,D6,Dc).
mossa_v(Prec,cg12,[A,B,V3,C,V5,[V6,D6],[V7,D7],D],  [A,B,V5,C,V3,[V7,Db],[V6,Da],D]):-
	Prec\==cg12, cambia_dir(xy,D6,Da), cambia_dir(xy,D7,Db).
% Il seguente gruppo da tre non  necessario, ma selezionarlo aiuta a risparmiare quelche mossa; stesso tempo.
%mossa_v(Prec,cr12,[A,V2,B,C,D,V6,[V7,D7],[V8,D8]],  [A,V6,B,C,D,V2,[V8,Dd],[V7,Da]]):- 
%	Prec\==cr12, cambia_dir(xz,D7,Da), cambia_dir(xz,D8,Dd).
%mossa_v(Prec,gc12,[A,V2,B,C,D,[V6,D6],[V7,D7],V8],  [A,V8,B,C,D,[V7,Dd],[V6,Da],V2]):-
%	Prec\==gc12, cambia_dir(xy,D6,Da), cambia_dir(xy,D7,Dd).
%mossa_v(Prec,rg12,[A,[V2,D2],B,C,D,V6,[V7,D7],V8],  [A,[V7,Dd],B,C,D,V8,[V2,Da],V6]):-
%	Prec\==rg12, cambia_dir(yz,D2,Da), cambia_dir(yz,D7,Dd).

%mossa_v(Prec,vr12,[[V1,D1],A,B,C,V5,D,V7,[V8,D8]],  [[V8,Dd],A,B,C,V7,D,V5,[V1,Da]]):-
%	Prec\==vr12, cambia_dir(yz,D1,Da), cambia_dir(yz,D8,Dd).
%mossa_v(Prec,rc12,[V1,A,B,C,V5,D,[V7,D7],[V8,D8]],  [V5,A,B,C,V1,D,[V8,Dd],[V7,Da]]):-
%	Prec\==rc12, cambia_dir(xz,D7,Da), cambia_dir(xz,D8,Dd).
%mossa_v(Prec,cv12,[V1,A,B,C,[V5,D5],D,V7,[V8,D8]],  [V7,A,B,C,[V8,Dd],D,V1,[V5,Da]]):- 
%	Prec\==cv12, cambia_dir(yz,D5,Da), cambia_dir(yz,D8,Dd).

% Sequenze di 16 mosse che orientano un vertice e cambiano orientamento ad un altro.
% Sono a coppie: una ruota in senso orario, l'altra in senso antiorario.
% Vengono provate una sola volta ed in ordine. L'ultima (gr16) sistema entrambi gli ultimi 2 vertici,  infatti uno stato
% non valido avere un solo vertice non  orientato.
mossa_v_o(bv16,1,[[V1,z],V2,V3,[V4,D4],V5,V6,V7,V8],  [[V1,x],V2,V3,[V4,Db],V5,V6,V7,V8]):- %1
	ruota_orario(D4,Db). % Sequenza: [b,v2,b3,v,b3,v2,b,v2,b,v,b3,v3,b3,v,b,v2]
mossa_v_o(vb16,1,[[V1,y],V2,V3,[V4,D4],V5,V6,V7,V8],  [[V1,x],V2,V3,[V4,Db],V5,V6,V7,V8]):- %1
	ruota_antiorario(D4,Db). % Sequenza: [v,b2,v3,b,v3,b2,v,b2,v,b,v3,b3,v3,b,v,b2]
mossa_v_o(ba16,2,[V1,V2,[V3,D3],[V4,y],V5,V6,V7,V8],  [V1,V2,[V3,Da],[V4,x],V5,V6,V7,V8]):- %2
	ruota_antiorario(D3,Da). % Sequenza: [b,a2,b3,a,b3,a2,b,a2,b,a,b3,a3,b3,a,b,a2]
mossa_v_o(ab16,2,[V1,V2,[V3,D3],[V4,z],V5,V6,V7,V8],  [V1,V2,[V3,Da],[V4,x],V5,V6,V7,V8]):- %2
	ruota_orario(D3,Da). % Sequenza: [a,b2,a3,b,a3,b2,a,b2,a,b,a3,b3,a3,b,a,b2]
mossa_v_o(ga16,3,[V1,V2,[V3,z],V4,V5,[V6,D6],V7,V8],  [V1,V2,[V3,x],V4,V5,[V6,Db],V7,V8]):- %3
	ruota_orario(D6,Db). % Sequenza: [g,a2,g3,a,g3,a2,g,a2,g,a,g3,a3,g3,a,g,a2]
mossa_v_o(ag16,3,[V1,V2,[V3,y],V4,V5,[V6,D6],V7,V8],  [V1,V2,[V3,x],V4,V5,[V6,Db],V7,V8]):- %3
	ruota_antiorario(D6,Db). % Sequenza: [a,g2,a3,g,a3,g2,a,g2,a,g,a3,g3,a3,g,a,g2]
mossa_v_o(ca16,4,[V1,V2,V3,V4,[V5,D5],[V6,y],V7,V8],  [V1,V2,V3,V4,[V5,Da],[V6,x],V7,V8]):- %4
	ruota_antiorario(D5,Da). % Sequenza: [c,a2,c3,a,c3,a2,c,a2,c,a,c3,a3,c3,a,c,a2]
mossa_v_o(ac16,4,[V1,V2,V3,V4,[V5,D5],[V6,z],V7,V8],  [V1,V2,V3,V4,[V5,Da],[V6,x],V7,V8]):- %4
	ruota_orario(D5,Da). % Sequenza: [a,c2,a3,c,a3,c2,a,c2,a,c,a3,c3,a3,c,a,c2]
mossa_v_o(cv16,5,[V1,V2,V3,V4,[V5,z],V6,V7,[V8,D8]],  [V1,V2,V3,V4,[V5,x],V6,V7,[V8,Db]]):- %5
	ruota_orario(D8,Db). % Sequenza: [c,v2,c3,v,c3,v2,c,v2,c,v,c3,v3,c3,v,c,v2]
mossa_v_o(vc16,5,[V1,V2,V3,V4,[V5,y],V6,V7,[V8,D8]],  [V1,V2,V3,V4,[V5,x],V6,V7,[V8,Db]]):- %5
	ruota_antiorario(D8,Db). % Sequenza: [v,c2,v3,c,v3,c2,v,c2,v,c,v3,c3,v3,c,v,c2]
mossa_v_o(cr16,6,[V1,V2,V3,V4,V5,V6,[V7,D7],[V8,y]],  [V1,V2,V3,V4,V5,V6,[V7,Da],[V8,x]]):- %6
	ruota_antiorario(D7,Da). % Sequenza: [c,r2,c3,r,c3,r2,c,r2,c,r,c3,r3,c3,r,c,r2]
mossa_v_o(rc16,6,[V1,V2,V3,V4,V5,V6,[V7,D7],[V8,z]],  [V1,V2,V3,V4,V5,V6,[V7,Da],[V8,x]]):- %6
	ruota_orario(D7,Da). % Sequenza: [r,c2,r3,c,r3,c2,r,c2,r,c,r3,c3,r3,c,r,c2]
mossa_v_o(gr16,7,[V1,[V2,D2],V3,V4,V5,V6,[V7,z],V8],  [V1,[V2,Da],V3,V4,V5,V6,[V7,x],V8]):- %7
	ruota_orario(D2,Da). % Sequenza: [g,r2,g3,r,g3,r2,g,r2,g,r,g3,r3,g3,r,g,r2]
mossa_v_o(gr16,7,[V1,[V2,D2],V3,V4,V5,V6,[V7,y],V8],  [V1,[V2,Da],V3,V4,V5,V6,[V7,x],V8]):- %7
	ruota_antiorario(D2,Da). % Sequenza: [r,g2,r3,g,r3,g2,r,g2,r,g,r3,g3,r3,g,r,g2]
%mossa_v_o(br16,8,[[V1,D1],[V2,D2],V3,V4,V5,V6,V7,V8],  [[V1,Da],[V2,Db],V3,V4,V5,V6,V7,V8]):- %8
%	ruota_antiorario(D1,Da), ruota_antiorario(D2,Db). % Sequenza: [b,r2,b3,r,b3,r2,b,r2,b,r,b3,r3,b3,r,b,r2]

% RUOTA_ORARIO e RUOTA_ANTIORARIO
ruota_orario(x,y):- !.
ruota_orario(y,z):- !.
ruota_orario(z,x).
ruota_antiorario(y,x):- !.
ruota_antiorario(z,y):- !.
ruota_antiorario(x,z).

% CAMBIA_DIR
cambia_dir(xy,z,z):- !. 
cambia_dir(xy,x,y):- !.
cambia_dir(xy,y,x).
cambia_dir(xz,y,y):- !.
cambia_dir(xz,x,z):- !.
cambia_dir(xz,z,x).
cambia_dir(yz,x,x):- !.
cambia_dir(yz,y,z):- !.
cambia_dir(yz,z,y).

% INIZIALIZZAZIONE
vertici_ko([[cag,z],[bvr,y],[cgr,y],[bav,z],[crv,z],[bga,y],[cva,y],[brg,z]]).
spigoli_ko([[ca,y],[gr,z],[vr,x],[bg,x],[br,y],[cg,y],[ba,y],[cv,y],[ga,x],[bv,y],[va,z],[cr,x]]).
% b,v2,r,g2,a,c,b3,v3,r3,a3,c2.	% Per aggiustarli: c2,a,2,v,b,c3,a3,g2,r3,v2,b3.

% GO
go:- vertici_ko(V), trova_vertici(V,8,1), nl, write('Finalmente'), nl.

% TROVA_VERTICI	
trova_vertici(Vertici,0,_):- !, nl, nl, orientamento_vertici(Vertici,1).
trova_vertici(V,Diversi,Profondit_max):- posizione_vertici(V,Sequenza,z,_,Diversi,Profondit_max,0), !,
	conta_mosse(Sequenza,Numero_mosse), calcola_stato(V,Sequenza,Vn), vertici_fuori_posto(Vn,FP,D), 
	write('Dopo '), write(Numero_mosse), write(' mosse ('),	write(Sequenza), write(') ,'),
	write(D), write(' vertici fuori posto:'), nl, write(FP), nl, write(Vn), nl,
	vertici_ok(V_ok), write(V_ok), write('  il goal.'), nl, trova_vertici(Vn,D,1).
trova_vertici(V,Diversi,Profondit_max):- 
	P is Profondit_max+1, write('profondit: '), Pr is P-1, write(Pr),nl, trova_vertici(V,Diversi,P).

% POSIZIONE_VERTICI
% Cerca di trovare una sequenza di mosse che porti ad avere un numero minore di "Diversi" vertici fuori posto.
posizione_vertici(V,[],_,singole_ok,Diversi,_,_):- n_vertici_fuori_posto(V,N), Diversi>N, !.	% ho trovato!
posizione_vertici(_,_,_,_,_,Profondit_max,Profondit_max):- !, fail. % forzo il backtracking
posizione_vertici(Vertici,[Mossa|Sequenza],Prec,Singola,Diversi,Profondit_max,Profondit):- 
	mossa_v(Prec,Mossa,Vertici,Vn), P is Profondit+1,
	posizione_vertici(Vn,Sequenza,Mossa,Singola,Diversi,Profondit_max,P).
posizione_vertici(Vertici,[Mossa|Sequenza],_,Singola,Diversi,Profondit_max,Profondit):- var(Singola), !,
	mossa_vs(_,Mossa,Vertici,Vn), 				% Prima mossa singola
	P is Profondit+1, posizione_vertici(Vn,Sequenza,Mossa,Mossa,Diversi,Profondit_max,P).
posizione_vertici(Vertici,[Mossa|Sequenza],_,Singola,Diversi,Profondit_max,Profondit):-
	mossa_vs(Singola,Mossa,Vertici,Vn), 			% Seconda mossa singola
	P is Profondit+1, posizione_vertici(Vn,Sequenza,Mossa,_,Diversi,Profondit_max,P).

% CONTA_MOSSE
conta_mosse(Sequenza,Numero_mosse):- length(Sequenza,0,Numero_mosse).
length([],N,N).
length([_|T],Acc,Numero_mosse):- Acc1 is Acc+1, length(T,Acc1,Numero_mosse).

% CALCOLA_STATO
calcola_stato(V,[],V).
calcola_stato(V,[Mossa|Sequenza],Vn):- mossa_v(z,Mossa,V,V1), !, calcola_stato(V1,Sequenza,Vn).
calcola_stato(V,[Mossa|Sequenza],Vn):- mossa_vs(_,Mossa,V,V1), calcola_stato(V1,Sequenza,Vn).

% VERTICI_FUORI_POSTO
vertici_fuori_posto(Vertici,FP,N):- vertici_ok(V_ok), vertici_fuori_posto(Vertici,V_ok,FP,0,N).

vertici_fuori_posto([],[],[],N,N).
vertici_fuori_posto([[H,_]|T],[[H,_]|T1],L,Acc,N):-  !, vertici_fuori_posto(T,T1,L,Acc,N).
vertici_fuori_posto([[H,O]|T],[[H1,O1]|T1],[[[H,O],[H1,O1]]|L],Acc,N):- 
	Acc1 is Acc+1, vertici_fuori_posto(T,T1,L,Acc1,N).

% N_VERTICI_FUORI_POSTO
n_vertici_fuori_posto(Vertici,N):- vertici_ok(V_ok), n_vertici_fuori_posto(Vertici,V_ok,0,N).

n_vertici_fuori_posto([],[],N,N).
n_vertici_fuori_posto([[H,_]|T],[[H,_]|T1],Acc,N):- !, n_vertici_fuori_posto(T,T1,Acc,N).
n_vertici_fuori_posto([_|T],[_|T1],Acc,N):- Acc1 is Acc+1, n_vertici_fuori_posto(T,T1,Acc1,N).

% ORIENTAMENTO_VERTICI
orientamento_vertici(Vertici,8):- not(vertici_ok(Vertici)), nl, write('Configurazione non corretta.'), nl, !, fail.
orientamento_vertici(Vertici,8):- !, nl, write('Vertici tutti a posto:'), nl, write(Vertici), nl.
orientamento_vertici(Vertici,N):- mossa_v_o(Mossa,N,Vertici,Vn), !, write(N), write(' vertici orientati, con '),
	write(Mossa), write(':'), nl, write(Vertici), write(' prima,'), nl, write(Vn), write(' dopo.'),nl,
	N1 is N+1, orientamento_vertici(Vn,N1).
orientamento_vertici(Vertici,N):- N1 is N+1, orientamento_vertici(Vertici,N1).

% ORIENTAMENTO_VERTICI:	Per orientare i vertici non secondo un ordine prestabilito.
%orientamento_vertici(_,0,_):- not(vertici_ok(Vertici)), nl, write('Configurazione non corretta.'), nl, !, fail.
%orientamento_vertici(_,0,_):- !, nl, write('Finalmente'), nl.
%orientamento_vertici(V,Diversi,Profondit_max):- orienta_vertici(V,Sequenza,Diversi,Profondit_max,0), !,
%	conta_mosse(Sequenza,Numero_mosse), vedi_stato(V,Sequenza,Vn), vertici_non_orientati(Vn,NO,D), 
%	write('Dopo '), write(Numero_mosse), write(' mosse ('), write(Sequenza), write(') ,'),
%	write(D), write(' vertici non orientati.'), nl, write(NO), nl, write(Vn), nl,
%	vertici_ok(V_ok), write(V_ok), write('  il goal.'), nl, orientamento_vertici(Vn,D,1).
%orientamento_vertici(V,Diversi,Profondit_max):- 
%	P is Profondit_max+1, write('profondit: '), Pr is P-1, write(Pr),nl, orientamento_vertici(V,Diversi,P).

% ORIENTA_VERTICI
% Cerca di trovare una sequenza di mosse che porti ad avere un numero minore di "Diversi" vertici mal orientati.
%orienta_vertici(V,[],Diversi,_,_):-	n_vertici_non_orientati(V,N), Diversi>N, !.		% ho trovato!
%orienta_vertici(_,_,_,Profondit_max,Profondit_max):- !, fail. % forzo il backtracking
%orienta_vertici(Vertici,[Mossa|Sequenza],Diversi,Profondit_max,Profondit):- mossa_v_o(Mossa,Vertici,Vn),
%	P is Profondit+1, orienta_vertici(Vn,Sequenza,Diversi,Profondit_max,P).

% VEDI_STATO
vedi_stato(V,[],V).
vedi_stato(V,[Mossa|Sequenza],Vn):- mossa_v_o(Mossa,V,V1), vedi_stato(V1,Sequenza,Vn).

% VERTICI_NON_ORIENTATI
vertici_non_orientati(Vertici,NO,N):- vertici_ok(V_ok), vertici_non_orientati(Vertici,V_ok,NO,0,N).

vertici_non_orientati([],[],[],N,N).
vertici_non_orientati([[H,O]|T],[[H,O]|T1],L,Acc,N):-  !, vertici_non_orientati(T,T1,L,Acc,N).
vertici_non_orientati([[H,O]|T],[[H1,O1]|T1],[[[H,O],[H1,O1]]|L],Acc,N):- 
	Acc1 is Acc+1, vertici_non_orientati(T,T1,L,Acc1,N).

% N_VERTICI_NON_ORIENTATI
n_vertici_non_orientati(Vertici,N):- vertici_ok(V_ok), n_vertici_non_orientati(Vertici,V_ok,0,N).

n_vertici_non_orientati([],[],N,N).
n_vertici_non_orientati([[H,O]|T],[[H,O]|T1],Acc,N):- !, n_vertici_non_orientati(T,T1,Acc,N).
n_vertici_non_orientati([_|T],[_|T1],Acc,N):- Acc1 is Acc+1, n_vertici_non_orientati(T,T1,Acc1,N).

not(X):- call(X),!,fail.
not(_).
